@@ -2,10 +2,6 @@
 #
 # Authors: F.Mertens
 
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-
 import re
 import tables
 import warnings
@@ -143,14 +139,7 @@ class WindowFunction(Mask):
         image_meta.set('PEWINFCT', '%s_%s' % (self.name, self.circular))
 
     def generate_window(self, nx):
-        w = get_window(self.name, nx)
-        if self.circular:
-            m = (nx - 1) // 2
-            x = np.linspace(-m, m, nx)
-            xx, yy = np.meshgrid(x, x)
-            return np.interp(np.sqrt(xx ** 2 + yy ** 2), x, w)
-        else:
-            return np.outer(w, w)
+        return psutil.generate_2d_window(self.name, nx, circular=self.circular)
 
     def generate(self, meta_data):
         return self.generate_window(meta_data.shape[0])
@@ -158,6 +147,8 @@ class WindowFunction(Mask):
 
 class BasePrimaryBeam(Mask):
 
+    name = 'pb'
+
     def __init__(self, masks, freq=None):
         self.freq = freq
         Mask.__init__(self, masks)
@@ -204,13 +195,7 @@ class PrimaryBeam(BasePrimaryBeam):
         return self.alpha_tapering * lamb / self.antenna_diameter
 
     def generate_beam(self, fwhm, res, shape):
-        nx, ny = shape
-
-        thxval = res * np.arange(-nx / 2., nx / 2.)
-        thyval = res * np.arange(-ny / 2., ny / 2.)
-        thx, thy = np.meshgrid(thxval, thyval)
-
-        return psutil.get_beam(np.sqrt(thx ** 2 + thy ** 2), self.beam_type, fwhm, None)
+        return psutil.get_beam_cart(res, tuple(shape), self.beam_type, fwhm, n_sidelobe=None)
 
     def generate(self, meta_data, freq=None):
         fwhm = self.get_fwhm(freq=freq)
@@ -392,6 +377,9 @@ class ImageMetaData(object):
         if name in self:
             del self.kargs[name]
 
+    def update(self, d):
+        self.kargs.update(d)
+
     @property
     def res(self):
         return abs(np.radians(self.wcs.wcs.cdelt[0]))
@@ -761,6 +749,9 @@ class DataCubeCombiner(object):
             self.total_time += cube.meta.total_time
 
     def get(self, min_n_nights=None):
+        if self.cube is None:
+            return self.cube
+
         cube = self.cube.new_with_data(np.divide(self.cube.data, self.total_weights,
                                                  where=self.total_weights != 0), weights=self.cube.weights)
         cube.meta.set('PETOTTIM', self.total_time)
@@ -1392,7 +1383,7 @@ class CartDataCube(DataCube):
                 n_filt = np.sum(~idx_uv)
                 n_tot = float(len(self.uu) * len(self.freqs))
                 if verbose:
-                    print('Filtering %s visibilities (%.2f %%)' % (n_filt, n_filt // n_tot * 100))
+                    print('Filtering %s visibilities (%.2f %%)' % (n_filt, n_filt / n_tot * 100))
                 self.data[~idx_uv] = 0
                 self.weights.data[~idx_uv] = 0
             else:
@@ -1400,7 +1391,7 @@ class CartDataCube(DataCube):
                 n_filt = np.sum(~idx_uv)
                 n_tot = float(len(self.uu))
                 if verbose:
-                    print('Filtering %s modes (%.2f %%)' % (n_filt, n_filt // n_tot * 100))
+                    print('Filtering %s modes (%.2f %%)' % (n_filt, n_filt / n_tot * 100))
                 self.filter_uv_from_index(idx_uv)
 
     def filter_sefd_uv(self, max_sefd, min_sefd=0):
@@ -1656,6 +1647,27 @@ class MultiDataInfo(object):
         return self.d[night][2]
 
 
+class NoiseStdCube(CartDataCube):
+    
+    def __init__(self, data, uu, vv, freqs, meta, weights=None):
+        CartDataCube.__init__(self, data, uu, vv, freqs, meta, weights=weights)
+        
+    def new_with_data(self, data, cov_err=None, weights=None, freqs=None):
+        if freqs is None:
+            freqs = self.freqs
+
+        if weights is None:
+            weights = self.weights
+
+        assert data.shape[0] == len(freqs)
+        assert data.shape[1] == len(self.uu)
+
+        if weights is not None:
+            weights = weights.copy()
+
+        return NoiseStdCube(data, self.uu, self.vv, freqs, self.meta.copy(), weights=weights)
+
+
 class CartWeightCube(CartDataCube):
 
     def __init__(self, weight_cube, uu, vv, freqs, meta, uv_scale=None, freqs_n_nights=None):
@@ -1864,7 +1876,8 @@ class CartWeightCube(CartDataCube):
         total_time = self.meta.total_time
         f = time / total_time
 
-        noise_rms = sefd / (2 * df * int_time * f) ** 0.5
+        # noise_rms = sefd / (2 * df * int_time * f) ** 0.5
+        noise_rms = np.atleast_1d(sefd / (2 * df * int_time * f) ** 0.5)[:, None]
 
         # noise_rms for each complex part
         noise_rms = np.sqrt(0.5) * noise_rms
@@ -1910,6 +1923,31 @@ class CartWeightCube(CartDataCube):
 
         return noise_cube
 
+    def get_noise_std_cube(self, sefd, time):
+        df = self.meta.chan_width
+        int_time = self.meta.int_time
+        total_time = self.meta.total_time
+        f = time / total_time
+
+        noise_rms = np.atleast_1d(sefd / (2 * df * int_time * f) ** 0.5)
+
+        lamb = const.c.value / self.freqs
+        fov = (self.meta.shape[0] * self.meta.res) ** 2
+        jy2k = ((1e-26 * lamb ** 2) / (2 * const.k_B.value)) / fov
+        noise_rms = (jy2k * noise_rms)[:, None]
+
+        w = self.get() ** 0.5
+        noise_rms = noise_rms / w
+
+        weights = self.new_with_data(self.data * f)
+
+        noise_cube = NoiseStdCube(noise_rms, self.uu, self.vv, self.freqs.copy(),
+                                  self.meta.copy(), weights=weights)
+        noise_cube.meta.remove('PEWINFCT')
+        noise_cube.weights.meta.remove('PEWINFCT')
+
+        return noise_cube
+
     def new_with_data(self, data, cov_err=None, weights=None, freqs=None):
         """Create a new CartWeightCube using data and cov_err"""
         if freqs is None:
