@@ -17,6 +17,7 @@ import numpy as np
 from . import fitutil
 from . import datacube
 from . import psutil
+from . import sphcube
 
 
 class FitterResult(object):
@@ -189,7 +190,7 @@ class GprForegroundResult(FitterResult):
                 with open(os.path.join(dir_path, name + '.kern.json'), mode='r', encoding='utf-8') as f:
                     s = f.read()
                     kern = fitutil.GPy.kern.Kern.from_dict(json.loads(s))
-            except ValueError:
+            except (ValueError, TypeError):
                 print('Loading kernel parameter failed')
                 kern = None
 
@@ -279,7 +280,7 @@ class GprForegroundFit(AbstractForegroundFitter):
 
         cube_sub.set_weights(data_cube.weights)
 
-        return GprForegroundResult(gpr_res, cube_fit, cube_sub, cube_pre_fit, cube_post_fit, 
+        return GprForegroundResult(gpr_res, cube_fit, cube_sub, cube_pre_fit, cube_post_fit,
                                    cube_noise=data_cube_noise)
 
 
@@ -368,7 +369,12 @@ class MultiBaselinesGprForegroundFit(GprForegroundFit):
         self.du = du
 
     def _split_cube(self, cube, idx):
-        return datacube.CartDataCube(cube.data[:, idx], cube.uu[idx], cube.vv[idx], cube.freqs, cube.meta)
+        if isinstance(cube, datacube.CartDataCube):
+            return datacube.CartDataCube(cube.data[:, idx], cube.uu[idx], cube.vv[idx], cube.freqs, cube.meta)
+        elif isinstance(cube, sphcube.SphDataCube):
+            return sphcube.SphDataCube(cube.data[:, idx], cube.ll[idx], cube.mm[idx], cube.freqs, cube.meta)
+        else:
+            raise TypeError('Input cube should be a CartDataCube or a SphDataCube')
 
     def run(self, data_cube, data_cube_noise):
         multi_gpr_res = MultiGprForegroundResult(data_cube, data_cube_noise)
