@@ -87,16 +87,22 @@ class FixedFreqsFlagger(BaseFreqsFlagger):
         BaseFreqsFlagger.__init__(self, **kargs)
         # Frequencies list coma separated in MHz , e.g. 121.3,123.1
         self.add('freqs', '', str)
+        self.parse_dict(kargs)
 
     def get_outliers(self, i_cube, v_cube):
         fmhz = i_cube.freqs * 1e-6
         fwidth = psutil.robust_freq_width(i_cube.freqs) * 1e-6
+        fct_norm = lambda f: np.round((f - fmhz[0]) / fwidth).astype(int)
 
-        fmhz_outliers = np.array([float(k.strip()) for k in self.freqs.split(',')])
-
-        norm_fmhz = np.round((fmhz - fmhz[0]) / fwidth).astype(int)
-        norm_fmhz_outliers = np.round((fmhz_outliers - fmhz[0]) / fwidth).astype(int)
+        norm_fmhz = fct_norm(fmhz)
+        norm_fmhz_outliers = []
 
+        for s in self.freqs.split(','):
+            if '-' in s:
+                a, b = s.split('-')
+                norm_fmhz_outliers.extend(np.arange(fct_norm(float(a)), fct_norm(float(b))))
+            else:
+                norm_fmhz_outliers.append(fct_norm(float(s)))
         return np.in1d(norm_fmhz, norm_fmhz_outliers)
 
 
@@ -105,14 +111,16 @@ class UVDirectionFlagger(BaseUVFlagger):
     def __init__(self, **kargs):
         BaseUVFlagger.__init__(self, **kargs)
         self.add('direction_deg', 0, float)
+        self.add('extend', 0.8, float)
         self.parse_dict(kargs)
 
     def get_outliers(self, i_cube, v_cube):
-        du = 0.8 * np.median(np.diff(i_cube.uu))
+        du = self.extend * np.median(np.diff(i_cube.uu))
         d_rad = np.radians(self.direction_deg) % np.pi
         theta = np.arctan2(i_cube.uu, i_cube.vv) % np.pi
+        d = (theta - d_rad - np.pi / 2) % np.pi
 
-        uv_idx = (theta > (d_rad - np.arctan(du / i_cube.ru))) & (theta < (d_rad + np.arctan(du / i_cube.ru)))
+        uv_idx = (d > (np.pi / 2 - np.arctan(du / i_cube.ru))) & (d < (np.pi / 2 + np.arctan(du / i_cube.ru)))
 
         return uv_idx
 
@@ -134,14 +142,22 @@ class SigmaClipper(psutil.SimpleConfig):
     def get_sigma_clip_mask(self, x, y):
         med = np.nanmedian(y)
         rms = psutil.mad(y)
+        # plt.figure()
 
         if self.detrend_poly_deg > 0:
             mask = abs(y - med) < self.detrend_nsigma_clip * rms
             mw_fct = np.poly1d(np.polyfit(x[mask], y[mask], self.detrend_poly_deg))
+            # plt.scatter(x, y)
+            # plt.plot(x, mw_fct(x))
             y = y - mw_fct(x)
             med = np.nanmedian(y[mask])
             rms = psutil.mad(y[mask])
 
+        # plt.scatter(x, y)
+        # plt.axhline(med)
+        # plt.axhline(med + self.nsigma * rms)
+        # plt.savefig(self.name + '.png')
+
         return (y - med) > self.nsigma * rms
 
 
@@ -181,9 +197,10 @@ class FreqsWeightsFlagger(BaseFreqsFlagger):
 
     def get_outliers(self, i_cube, v_cube):
         weights = np.median(v_cube.weights.get(), axis=1)
+        mask = weights > 0
 
-        med = np.median(weights)
-        rms = psutil.mad(weights)
+        med = np.median(weights[mask])
+        rms = psutil.mad(weights[mask])
         mask = abs(weights - med) < 5 * rms
 
         mw_fct = np.poly1d(np.polyfit(i_cube.freqs[mask], weights[mask], self.trend_poly_deg))
@@ -311,7 +328,7 @@ class Flag(object):
 
 class FlaggerRunner(object):
 
-    def __init__(self, verbose=True):
+    def __init__(self, verbose=True, min_weights=2):
         self.flaggers = []
         self.original_i_cube = None
         self.original_v_cube = None
@@ -319,11 +336,16 @@ class FlaggerRunner(object):
         self.filtered_v_cube = None
         self.flag = None
         self.verbose = verbose
+        self.min_weights = min_weights
 
     def add(self, flagger):
         self.flaggers.append(flagger)
 
     def run(self, i_cube, v_cube):
+        i_cube_before_flagging = i_cube.copy()
+        if self.min_weights is not None:
+            i_cube.filter_min_weight(self.min_weights, verbose=self.verbose)
+            v_cube.filter_min_weight(self.min_weights, verbose=self.verbose)
         self.original_i_cube = i_cube.copy()
         self.original_v_cube = v_cube.copy()
         for flagger in self.flaggers:
@@ -332,7 +354,7 @@ class FlaggerRunner(object):
         self.filtered_i_cube = i_cube
         self.filtered_v_cube = v_cube
 
-        idx_freqs, idx_uv, _, _ = datacube.get_common_idx(self.original_i_cube, self.filtered_i_cube)
+        idx_freqs, idx_uv, _, _ = datacube.get_common_idx(i_cube_before_flagging, self.filtered_i_cube)
 
         if self.filtered_i_cube.weights is not None:
             idx_uv_zero_weights = (self.filtered_i_cube.weights.get().sum(axis=0) == 0)
@@ -341,11 +363,11 @@ class FlaggerRunner(object):
             idx_uv_zero_weights = None
             idx_freqs_zero_weights = None
 
-        if isinstance(self.original_i_cube, datacube.CartDataCube):
-            self.flag = Flag(self.original_i_cube.freqs, self.original_i_cube.uu, self.original_i_cube.vv,
+        if isinstance(i_cube_before_flagging, datacube.CartDataCube):
+            self.flag = Flag(i_cube_before_flagging.freqs, i_cube_before_flagging.uu, i_cube_before_flagging.vv,
                              idx_uv, idx_uv_zero_weights, idx_freqs, idx_freqs_zero_weights)
-        elif isinstance(self.original_i_cube, sphcube.SphDataCube):
-            self.flag = Flag(self.original_i_cube.freqs, self.original_i_cube.ll, self.original_i_cube.mm,
+        elif isinstance(i_cube_before_flagging, sphcube.SphDataCube):
+            self.flag = Flag(i_cube_before_flagging.freqs, i_cube_before_flagging.ll, i_cube_before_flagging.mm,
                              idx_uv, idx_uv_zero_weights, idx_freqs, idx_freqs_zero_weights)
         else:
             raise ValueError('Cube is not of a supported format')
@@ -355,8 +377,8 @@ class FlaggerRunner(object):
     def apply_last(self, cube):
         return self.flag.apply(cube)
 
-    def plot(self, figsize=(10, 8), **fig_kargs):
-        fig, ((ax1, ax2), (ax3, ax4)) = plt.subplots(ncols=2, nrows=2, figsize=figsize, **fig_kargs)
+    def plot(self, figsize=(10, 12), **fig_kargs):
+        fig, ((ax1, ax2), (ax3, ax4), (ax5, ax6)) = plt.subplots(ncols=2, nrows=3, figsize=figsize, **fig_kargs)
 
         idx1, idx1_uv, _, _ = datacube.get_common_idx(self.original_i_cube, self.filtered_i_cube)
         freqs_idx_filtered = ~idx1
@@ -366,8 +388,15 @@ class FlaggerRunner(object):
             freqs_idx_zero_w = self.filtered_i_cube.weights.data.sum(axis=1) == 0
             uv_idx_zero_w = self.filtered_i_cube.weights.data.sum(axis=0) == 0
 
-        ax1.plot(self.original_i_cube.freqs * 1e-6, self.original_i_cube.data.var(axis=1))
-        ax1.plot(self.original_i_cube.freqs * 1e-6, self.original_v_cube.data.var(axis=1))
+        ax1.plot(self.original_i_cube.freqs * 1e-6, self.original_i_cube.data.var(axis=1),
+                 c=psutil.lblue, label='I before')
+        ax1.plot(self.original_i_cube.freqs * 1e-6, self.original_v_cube.data.var(axis=1),
+                 c=psutil.lorange, label='V before')
+        ax1.plot(self.filtered_i_cube.freqs * 1e-6, self.filtered_i_cube.data.var(axis=1),
+                 c=psutil.dblue, label='I after')
+        ax1.plot(self.filtered_v_cube.freqs * 1e-6, self.filtered_v_cube.data.var(axis=1),
+                 c=psutil.dorange, label='V after')
+
         ax1.plot(self.original_i_cube.freqs[freqs_idx_filtered] * 1e-6,
                  self.original_v_cube.data.var(axis=1)[freqs_idx_filtered],
                  marker='x', ls='', c=psutil.black)
@@ -379,30 +408,53 @@ class FlaggerRunner(object):
 
         ax1.set_yscale('log')
         ax1.set_xlabel('Freqs [MHz]')
-        ax1.set_ylabel('Variance [Unnormalized]')
+        ax1.set_ylabel('Variance [Unormalized]')
+        ax1.legend()
+
+        ax2.plot(self.original_i_cube.freqs * 1e-6, self.original_v_cube.make_diff_cube_interp().estimate_freqs_sefd(),
+                 c=psutil.lorange, label='dV before')
+        ax2.plot(self.original_i_cube.freqs * 1e-6, self.original_i_cube.make_diff_cube_interp().estimate_freqs_sefd(),
+                 c=psutil.lblue, label='dI before')
+        ax2.plot(self.filtered_v_cube.freqs * 1e-6, self.filtered_v_cube.make_diff_cube_interp().estimate_freqs_sefd(),
+                 c=psutil.lorange, label='dV after')
+        ax2.plot(self.filtered_i_cube.freqs * 1e-6, self.filtered_i_cube.make_diff_cube_interp().estimate_freqs_sefd(),
+                 c=psutil.dblue, label='dI after')
+
+        ax2.plot(self.original_i_cube.freqs[freqs_idx_filtered] * 1e-6,
+                 self.original_v_cube.make_diff_cube_interp().estimate_freqs_sefd()[freqs_idx_filtered],
+                 marker='x', ls='', c=psutil.black)
+
+        ax2.set_yscale('log')
+        ax2.set_xlabel('Freqs [MHz]')
+        ax2.set_ylabel('SEFD [Unormalized]')
+        ax2.legend()
 
         if isinstance(self.original_i_cube, datacube.CartDataCube):
-            self.original_v_cube.weights.plot_uv(ax=ax2)
-            ax2.set_title('Weights')
-            self.original_v_cube.estimate_uv_sefd().plot_uv(ax=ax3)
-            ax3.set_title('SEFD V')
-            self.original_i_cube.make_diff_cube().estimate_uv_sefd().plot_uv(ax=ax4)
-            ax4.set_title('SEFD d_nu I')
+            self.original_v_cube.weights.plot_uv(ax=ax3)
+            ax3.set_title('Weights')
+            self.original_v_cube.estimate_uv_sefd().plot_uv(ax=ax4)
+            ax4.set_title('SEFD V')
+            self.original_i_cube.make_diff_cube().estimate_uv_sefd().plot_uv(ax=ax5)
+            ax5.set_title('SEFD d_nu I')
+            self.original_v_cube.make_diff_cube().estimate_uv_sefd().plot_uv(ax=ax6)
+            ax6.set_title('SEFD d_nu V')
 
-            for ax in [ax2, ax3, ax4]:
+            for ax in [ax3, ax4, ax5, ax6]:
                 ax.scatter(self.original_i_cube.uu[uv_idx_filtered],
                            self.original_i_cube.vv[uv_idx_filtered], c=psutil.black, s=40, marker='x')
                 if self.filtered_i_cube.weights is not None:
                     ax.scatter(self.filtered_i_cube.uu[uv_idx_zero_w],
                                self.filtered_i_cube.vv[uv_idx_zero_w], c=psutil.black, s=40, marker='+')
         elif isinstance(self.original_i_cube, sphcube.SphDataCube):
-            self.original_i_cube.plot_lm(ax=ax2, action_fct=np.mean)
-            ax2.set_title('Mean Stokes I')
-            self.original_v_cube.plot_lm(ax=ax3, action_fct=np.var, norm=LogNorm())
-            ax3.set_title('Var V')
-            self.original_i_cube.make_diff_cube().plot_lm(ax=ax4, action_fct=np.var, norm=LogNorm())
-            ax4.set_title('Var d_nu I')
-            for ax in [ax2, ax3, ax4]:
+            self.original_i_cube.plot_lm(ax=ax3, action_fct=np.mean)
+            ax3.set_title('Mean Stokes I')
+            self.original_v_cube.plot_lm(ax=ax4, action_fct=np.var, norm=LogNorm())
+            ax4.set_title('Var V')
+            self.original_i_cube.make_diff_cube().plot_lm(ax=ax5, action_fct=np.var, norm=LogNorm())
+            ax5.set_title('Var d_nu I')
+            self.original_v_cube.make_diff_cube().plot_lm(ax=ax6, action_fct=np.var, norm=LogNorm())
+            ax6.set_title('Var d_nu I')
+            for ax in [ax3, ax4, ax5, ax6]:
                 ax.scatter(self.original_i_cube.ll[uv_idx_filtered],
                            self.original_i_cube.mm[uv_idx_filtered], c=psutil.black, s=40, marker='x')
 
