@@ -2,10 +2,7 @@
 #
 # Authors: F.Mertens
 
-from __future__ import division
-from __future__ import print_function
-from __future__ import absolute_import
-
+import operator
 import numpy as np
 
 import matplotlib as mpl
@@ -17,6 +14,8 @@ from scipy.signal import get_window
 
 import astropy.units as u
 
+import h5py
+
 from . import psutil
 from . import pscart
 from . import pssph
@@ -374,7 +373,6 @@ class BasePowerSpectra(object):
         self._setup_cache()
 
     def _setup_cache(self):
-        self.omega_cache = psutil.Cache(self.get_omega)
         self.ps_cov_cache = psutil.Cache(self._get_ps_cov_err)
         self.ps2d_cov_cache = psutil.Cache(self._get_ps2d_cov_err)
         self.ps3d_cov_cache = psutil.Cache(self._get_ps3d_cov_err)
@@ -605,24 +603,30 @@ class BasePowerSpectra(object):
 
     def delay_transform(self, data_cube):
         ''' Delay transform data_cube '''
-        weights = self.get_weights(data_cube, True)
-        window = self.get_window_fct(data_cube)
+        if isinstance(data_cube, datacube.NoiseStdCube):
+            delay, dft_cube = psutil.lssa_diag_cov(data_cube.freqs, data_cube.data ** 2, 
+                                                   M=self.eor.M, dx=self.config.df)
+            dft_cube = dft_cube ** .5
+        else:
+            weights = self.get_weights(data_cube, True)
+            window = self.get_window_fct(data_cube)
 
-        delay, dft_cube = psutil.delay_transform_cube(data_cube.freqs, data_cube.data, M=self.eor.M,
-                                                      method=self.ft_method, dx=self.config.df,
-                                                      window=window, weights=weights,
-                                                      rmean_axis=self.config.rmean_axis)
+            delay, dft_cube = psutil.delay_transform_cube(data_cube.freqs, data_cube.data, M=self.eor.M,
+                                                          method=self.ft_method, dx=self.config.df,
+                                                          window=window, weights=weights,
+                                                          rmean_axis=self.config.rmean_axis)
 
-        if self.config.filter_kpar_min is not None or self.config.filter_wedge_theta > 0:
-            k_par = psutil.delay_to_k(delay, self.z)
-            k_per = psutil.l_to_k(2 * np.pi * data_cube.ru, self.z)
+            if self.config.filter_kpar_min is not None or self.config.filter_wedge_theta > 0:
+                k_par = psutil.delay_to_k(delay, self.z)
+                k_per = psutil.l_to_k(2 * np.pi * data_cube.ru, self.z)
 
-            wedge_kpar = psutil.wedge_fct(np.radians(self.config.filter_wedge_theta), self.z, k_per)
-            if self.config.filter_kpar_min is not None:
-                wedge_kpar += self.config.filter_kpar_min
+                wedge_kpar = psutil.wedge_fct(np.radians(self.config.filter_wedge_theta), self.z, k_per)
+                if self.config.filter_kpar_min is not None:
+                    wedge_kpar += self.config.filter_kpar_min
+
+                wedge = np.array([abs(k_par) < w_b for w_b in wedge_kpar]).T
+                dft_cube[wedge] = np.nan
 
-            wedge = np.array([abs(k_par) < w_b for w_b in wedge_kpar]).T
-            dft_cube[wedge] = np.nan
         return delay, dft_cube
 
     def _handle_cov_err(self, ps, ps_err, ps_cov_fct, data_cube, with_cov_err, *args, **kargs):
@@ -670,7 +674,7 @@ class PowerSpectraCart(BasePowerSpectra):
 
     def get_ps_norm(self, data_cube):
         '''Normalization factor for spatial PS'''
-        return self.X ** 2 / self.omega_cache.get(data_cube.meta)
+        return self.X ** 2 / self.get_omega(data_cube.meta)
 
     def get_ps_err_norm(self, data_cube, with_pb=False):
         mask = datacube.WindowFunction.from_meta(data_cube.meta)
@@ -884,7 +888,7 @@ class PowerSpectraSph(BasePowerSpectra):
 
     def get_ps_norm(self, alm_cube):
         # Normalization factor for 1D (spacial/frequency) PS
-        return self.X ** 2 / self.omega_cache.get(alm_cube.meta)
+        return self.X ** 2 / self.get_omega(alm_cube.meta)
 
     def get_omega(self, cube_meta):
         '''Primary Beam normalization factor'''
@@ -1598,6 +1602,48 @@ class SphericalPowerSpectra(PowerSpectraMath):
         return SphericalPowerSpectra(data * 1e-6, err * 1e-6, kbins, k_mean, n_eff=n_eff, k_std=k_std)
 
 
+class FourPanelPsResults(object):
+
+    def __init__(self, ps_gen, kbins, figsize=(10, 8)):
+        self.ps_gen = ps_gen
+        self.kbins = kbins
+        self.fig, ((self.ax1, self.ax2), (self.ax3, self.ax4)) = plt.subplots(ncols=2, nrows=2, figsize=figsize)
+        self.lgd = None
+
+    def add_cube(self, cube, label, ps_gen=None, **kargs):
+        if ps_gen is None:
+            ps_gen = self.ps_gen
+        ps_gen.get_variance(cube).plot(ax=self.ax1, label=label, **kargs)
+        ps_gen.get_ps3d(self.kbins, cube).plot(ax=self.ax2, label=label, **kargs)
+        ps_gen.get_ps2d(cube).plot_kpar(ax=self.ax3, label=label, **kargs)
+        ps_gen.get_ps2d(cube).plot_kper(ax=self.ax4, label=label, **kargs)
+
+    def add_cube_ps_diff(self, cube1, cube2, label, ps_gen=None, **kargs):
+        if ps_gen is None:
+            ps_gen = self.ps_gen
+        (ps_gen.get_variance(cube1) - ps_gen.get_variance(cube2)).plot(ax=self.ax1, label=label, **kargs)
+        (ps_gen.get_ps2d(cube1) - ps_gen.get_ps2d(cube2)).plot_kpar(ax=self.ax3, label=label, **kargs)
+        (ps_gen.get_ps2d(cube1) - ps_gen.get_ps2d(cube2)).plot_kper(ax=self.ax4, label=label, **kargs)
+        ps_gen.get_ps3d_with_noise(self.kbins, cube1, cube2).plot(ax=self.ax2, label=label, **kargs)
+
+    def add_ps_stacker(self, ps_stacker, label, **kargs):
+        ps_stacker.get_variance().plot(ax=self.ax1, label=label, **kargs)
+        ps_stacker.get_ps3d().plot(ax=self.ax2, label=label, **kargs)
+        ps_stacker.get_ps2d().plot_kpar(ax=self.ax3, label=label, **kargs)
+        ps_stacker.get_ps2d().plot_kper(ax=self.ax4, label=label, **kargs)
+
+    def done(self, ncol_legend=3):
+        self.lgd = self.fig.legend(*self.ax1.get_legend_handles_labels(),
+                                   bbox_to_anchor=(0.5, 1.04), loc="upper center", ncol=ncol_legend)
+        self.fig.tight_layout()
+
+    def savefig(self, filename, **kargs):
+        if self.lgd is None:
+            print('Warning: run done() before savefig()')
+        else:
+            self.fig.savefig(filename, bbox_extra_artists=(self.lgd,), bbox_inches='tight')
+
+
 class MultiNight2DPowerSpectra(PowerSpectraMath):
 
     def __init__(self, ps, err, x, xlabel, y, ylabel):
@@ -1638,3 +1684,385 @@ class MultiNight2DPowerSpectra(PowerSpectraMath):
 
         ax.set_xlabel(self.xlabel)
         ax.set_ylabel(self.ylabel)
+
+
+class PsStacker(object):
+
+    def __init__(self, ps_gen, kbins):
+        self.all_ps3d = []
+        self.all_ps = []
+        self.all_ps2d = []
+        self.all_var = []
+        self.ps_gen = ps_gen
+        self.kbins = kbins
+
+    def add(self, cube):
+        self.all_ps3d.append(self.ps_gen.get_ps3d(self.kbins, cube))
+        self.all_ps.append(self.ps_gen.get_ps(cube))
+        self.all_ps2d.append(self.ps_gen.get_ps2d(cube))
+        self.all_var.append(self.ps_gen.get_variance(cube))
+
+    def get_ps(self):
+        return SpatialPowerSpectraMC(self.all_ps)
+
+    def get_ps2d(self):
+        return CylindricalPowerSpectraMC(self.all_ps2d)
+
+    def get_variance(self):
+        return VarianceMC(self.all_var)
+
+    def get_ps3d(self):
+        return SphericalPowerSpectraMC(self.all_ps3d)
+
+
+
+class PowerSpectraMC(object):
+
+    def __init__(self, all_ps):
+        self.all_ps = all_ps
+        self.med = np.median([ps.data for ps in self.all_ps], axis=0)
+        self.std = psutil.robust_std([ps.data for ps in self.all_ps], axis=0)
+        self.q16 = np.quantile(np.array([ps.data for ps in self.all_ps]), .16, axis=0)
+        self.q84 = np.quantile(np.array([ps.data for ps in self.all_ps]), .84, axis=0)
+        self.q2_5 = np.quantile(np.array([ps.data for ps in self.all_ps]), .025, axis=0)
+        self.q97_5 = np.quantile(np.array([ps.data for ps in self.all_ps]), .975, axis=0)
+        self.p0 = self.all_ps[0]
+
+    def __operation__(self, other, op):
+        all_ps = []
+        for ps in self.all_ps:
+            if isinstance(other, PowerSpectraMC):
+                all_ps.append(op(ps, np.random.choice(other.all_ps)))
+            else:
+                all_ps.append(op(ps, other.new_with_data(np.random.normal(other.data, other.err), other.err)))
+
+        return self.__class__(all_ps)
+
+    def __add__(self, other):
+        return self.__operation__(other, operator.add)
+
+    def __sub__(self, other):
+        return self.__operation__(other, operator.sub)
+
+    def __mul__(self, other):
+        if psutil.is_number(other):
+            return self.__class__([k.new_with_data(other * k.data, other * k.err, other * k.w) for k in self.all_ps])
+        else:
+            return self.__operation__(other, operator.mul)
+
+    def __rmul__(self, other):
+        return self.__mul__(other)
+
+    def __div__(self, other):
+        return self.__truediv__(other)
+
+    def __truediv__(self, other):
+        return self.__operation__(other, operator.truediv)
+
+    def get_data(self, mkelvin=False):
+        a = 1
+        temp_unit = 'K'
+        if mkelvin:
+            a = 1e6
+            temp_unit = 'mK'
+
+        return a * self.med, a * self.q16, a * self.q84, a * self.q2_5, a * self.q97_5, temp_unit
+
+
+
+class SphericalPowerSpectraMC(PowerSpectraMC, SphericalPowerSpectra):
+
+    def __init__(self, ps3d_all):
+        PowerSpectraMC.__init__(self, ps3d_all)
+        SphericalPowerSpectra.__init__(self, self.med, self.std, self.p0.k_bins, self.p0.k_mean,
+                                             n_eff=self.p0.n_eff, k_std=self.p0.k_std,
+                                             ps3d_q16=self.q16, ps3d_q84=self.q84)
+
+    def get_upper(self, nsigma=2, mkelvin=True, kelvin_square=False):
+        if nsigma == 1:
+            u = self.q84
+        elif nsigma == 2:
+            u = self.q97_5
+        else:
+            raise ValueError('nsigma must be 1 or 2')
+        if mkelvin:
+            u = u * 1e6
+        if not kelvin_square:
+            u = np.sqrt(u)
+
+        return u
+
+    def plot(self, ax=None, show68=True, show95=True, marker='+', mkelvin=True,
+             title=None, kerr_as_kbins=False, **kargs):
+        if ax is None:
+            fig, ax = plt.subplots()
+
+        d, q16, q84, q2_5, q97_5, temp_unit = self.get_data(mkelvin=mkelvin)
+
+        if show68:
+            ax.fill_between(self.k_mean, q16, q84, alpha=kargs.get('alpha', 0.5), color=kargs.get('c', None))
+        if show95:
+            ax.fill_between(self.k_mean, q2_5, q97_5, alpha=0.5 * kargs.get('alpha', 0.5), color=kargs.get('c', None))
+
+        ax.plot(self.k_mean, d, marker=marker, **kargs)
+
+        ax.set_yscale('log', **nonpos_arg)
+        ax.set_xscale('log')
+
+        ax.set_ylabel(r'$\Delta^2 (k)\,[\mathrm{%s}]$' % temp_unit)
+        ax.set_xlabel(r'$k\,[\mathrm{h\,cMpc^{-1}}]$')
+
+        ax.set_xlim(self.k_bins.min(), self.k_bins.max())
+
+        if title is not None:
+            ax.set_title(title)
+
+    def save(self, filename):
+        data = [p.data for p in self.all_ps]
+        err = [p.err for p in self.all_ps]
+        a = [self.k_bins, self.k_mean, self.k_std, data, err, self.n_eff]
+        with h5py.File(filename, 'w') as hf:
+            for d, l in zip(a, ['k_bins', 'k_mean', 'k_std', 'data', 'err', 'n_eff']):
+                if d is not None:
+                    hf.create_dataset(l, data=d)
+
+    @staticmethod
+    def load(filename):
+        with h5py.File(filename, 'r') as hf:
+            data = hf.get('data')[:]
+            err = hf.get('err')[:]
+            k_bins = hf.get('k_bins')[:]
+            k_mean = hf.get('k_mean')[:]
+            k_std = hf.get('k_std', default=None)
+            n_eff = hf.get('n_eff', default=None)
+        all_ps = [SphericalPowerSpectra(d, e, k_bins, k_mean, n_eff=n_eff, k_std=k_std)
+                  for d, e in zip(data, err)]
+
+        return SphericalPowerSpectraMC(all_ps)
+
+
+class CylindricalPowerSpectraMC(PowerSpectraMC, CylindricalPowerSpectra):
+
+    def __init__(self, all_ps2d):
+        PowerSpectraMC.__init__(self, all_ps2d)
+        CylindricalPowerSpectra.__init__(self, self.med, self.std, self.p0.delay, self.p0.el, self.p0.k_per, 
+                                               self.p0.k_par, n_eff=self.p0.n_eff, ps2d_w=self.p0.w)
+
+    def plot_kpar(self, ax=None, show68=True, show95=True, delay=False, weighted=True, mkelvin=False, **kargs):
+        if ax is None:
+            fig, ax = plt.subplots()
+
+        d, q16, q84, q2_5, q97_5, temp_unit = self.get_data(mkelvin=mkelvin)
+
+        if weighted and isinstance(self.w, np.ndarray):
+            y, q16, q84, q2_5, q97_5 = psutil.nanaverage(np.array([d, q16, q84, q2_5, q97_5]),
+                                                         self.w[None], axis=2)
+        else:
+            y, q16, q84, q2_5, q97_5 = np.nanmean([d, q16, q84, q2_5, q97_5], axis=2)
+
+        if delay:
+            x = self.delay
+        else:
+            x = self.k_par
+
+        if show68:
+            ax.fill_between(x, q16, q84, alpha=kargs.get('alpha', 0.5), color=kargs.get('c', None))
+        if show95:
+            ax.fill_between(x, q2_5, q97_5, alpha=0.5 * kargs.get('alpha', 0.25), color=kargs.get('c', None))
+
+        ax.plot(x, y, **kargs)
+
+        ax.set_yscale('log', **nonpos_arg)
+
+        if delay:
+            ax.set_xlabel('Delay (us)')
+        else:
+            ax.set_xlabel(r'$k_{\parallel}\,[\mathrm{h\,cMpc^{-1}}]$')
+
+        ax.set_ylabel(r'$P(k_{\parallel})\,[\mathrm{%s^2\,h^{-3}\,cMpc^3}]$' % temp_unit)
+
+    def plot_kper(self, ax=None, show68=True, show95=True, normalize=False, weighted=True, mkelvin=False, **kargs):
+        if ax is None:
+            fig, ax = plt.subplots()
+
+        d, q16, q84, q2_5, q97_5, temp_unit = self.get_data(mkelvin=mkelvin)
+
+        if weighted and isinstance(self.w, np.ndarray):
+            y, q16, q84, q2_5, q97_5 = psutil.nanaverage(np.array([d, q16, q84, q2_5, q97_5]),
+                                                         self.w[None], axis=1)
+        else:
+            y, q16, q84, q2_5, q97_5 = np.nanmean([d, q16, q84, q2_5, q97_5], axis=1)
+
+        if normalize:
+            n = self.k_per ** 2 / (2 * np.pi)
+            y, q16, q84, q2_5, q97_5 = [n * k for k in (y, q16, q84, q2_5, q97_5)]
+
+        x = self.k_per
+
+        if show68:
+            ax.fill_between(x, q16, q84, alpha=kargs.get('alpha', 0.5), color=kargs.get('c', None))
+        if show95:
+            ax.fill_between(x, q2_5, q97_5, alpha=0.5 * kargs.get('alpha', 0.25), color=kargs.get('c', None))
+
+        ax.plot(x, y, **kargs)
+
+        ax.set_yscale('log', **nonpos_arg)
+
+        ax.set_xlabel(r'$k_{\bot}\,[\mathrm{h\,cMpc^{-1}}]$')
+
+        if normalize:
+            ax.set_ylabel(r'$\Delta^2 (k_{\bot})\,[\mathrm{%s^2}]$' % temp_unit)
+        else:
+            ax.set_ylabel(r'$P(k_{\bot})\,[\mathrm{%s^2\,h^{-3}\,cMpc^3}]$' % temp_unit)
+
+    def save(self, filename):
+        data = [p.data for p in self.all_ps]
+        err = [p.err for p in self.all_ps]
+        a = [self.delay, self.el, self.k_per, self.k_par, data, err, self.w]
+        with h5py.File(filename, 'w') as hf:
+            for d, l in zip(a, ['delay', 'el', 'k_per', 'k_par', 'data', 'err', 'ps2d_w']):
+                hf.create_dataset(l, data=d)
+
+    @staticmethod
+    def load(filename):
+        with h5py.File(filename, 'r') as hf:
+            data = hf.get('data')[:]
+            err = hf.get('err')[:]
+            el = hf.get('el')[:]
+            delay = hf.get('delay')[:]
+            k_per = hf.get('k_per')[:]
+            k_par = hf.get('k_par')[:]
+            ps2d_w = hf.get('ps2d_w')
+            if not ps2d_w.size == 1:
+                ps2d_w = ps2d_w[:]
+        all_ps = [CylindricalPowerSpectra(d, e, delay, el, k_per, k_par, ps2d_w=ps2d_w)
+                  for d, e in zip(data, err)]
+
+        return CylindricalPowerSpectraMC(all_ps)
+
+
+class VarianceMC(PowerSpectraMC, Variance):
+
+    def __init__(self, all_var):
+        PowerSpectraMC.__init__(self, all_var)
+        Variance.__init__(self, self.med, self.std, self.p0.freqs, var_w=self.p0.w)
+
+    def plot(self, ax=None, show68=True, show95=True, mkelvin=True, title=None, **kargs):
+        if ax is None:
+            fig, ax = plt.subplots()
+
+        d, q16, q84, q2_5, q97_5, temp_unit = self.get_data(mkelvin=mkelvin)
+        f = self.freqs * 1e-6
+
+        if show68:
+            ax.fill_between(f, q16, q84, alpha=kargs.get('alpha', 0.5), color=kargs.get('c', None))
+        if show95:
+            ax.fill_between(f, q2_5, q97_5, alpha=0.5 * kargs.get('alpha', 0.25), color=kargs.get('c', None))
+
+        ax.plot(f, d, **kargs)
+
+        ax.set_yscale('log', **nonpos_arg)
+        ax.set_xlabel(r"$\mathrm{Frequency\,[MHz]}$")
+
+        ax.set_ylabel(r"$\mathrm{Variance\,[%s^2]}$" % temp_unit)
+
+        if title is not None:
+            ax.set_title(title)
+
+    def save(self, filename):
+        data = [p.data for p in self.all_ps]
+        err = [p.err for p in self.all_ps]
+        a = [self.freqs, data, err]
+        with h5py.File(filename, 'w') as hf:
+            for d, l in zip(a, ['freqs', 'data', 'err']):
+                hf.create_dataset(l, data=d)
+
+    @staticmethod
+    def load(filename):
+        with h5py.File(filename, 'r') as hf:
+            data = hf.get('data')[:]
+            err = hf.get('err')[:]
+            freqs = hf.get('freqs')[:]
+        all_ps = [Variance(d, e, freqs) for d, e in zip(data, err)]
+
+        return VarianceMC(all_ps)
+
+
+class SpatialPowerSpectraMC(PowerSpectraMC, SpatialPowerSpectra):
+
+    def __init__(self, all_var):
+        PowerSpectraMC.__init__(self, all_var)
+        SpatialPowerSpectra.__init__(self, self.med, self.std, self.p0.freqs, self.p0.el, 
+                                           self.p0.k_per, cl=self.p0.cl, n_eff=self.p0.n_eff, ps_w=self.p0.w)
+
+    def plot_kper(self, ax=None, show68=True, show95=True, normalize=False,
+                  weighted=True, l_lambda=False, mkelvin=True, **kargs):
+        if ax is None:
+            fig, ax = plt.subplots()
+
+        d, q16, q84, q2_5, q97_5, temp_unit = self.get_data(mkelvin=mkelvin)
+
+        if weighted and isinstance(self.w, np.ndarray):
+            y, q16, q84, q2_5, q97_5 = psutil.nanaverage(np.array([d, q16, q84, q2_5, q97_5]), self.w[None], axis=1)
+        else:
+            y, q16, q84, q2_5, q97_5 = np.nanmean([d, q16, q84, q2_5, q97_5], axis=1)
+
+        if normalize:
+            if self.cl:
+                n = self.el * (self.el + 1)
+            else:
+                n = self.k_per ** 2 / (2 * np.pi)
+            y, q16, q84, q2_5, q97_5 = [n * k for k in (y, q16, q84, q2_5, q97_5)]
+
+        if self.cl:
+            x = self.el
+            ax.set_xlabel(r'$\ell$')
+            if normalize:
+                ax.set_ylabel(r'$\ell (\ell + 1) C_{\ell}\,[\mathrm{%s}]$' % temp_unit)
+            else:
+                ax.set_ylabel(r'$C_{\ell}\,[\mathrm{%s}]$' % temp_unit)
+        else:
+            x = self.k_per
+            ax.set_xlabel(r'$k_{\bot}\,[\mathrm{h\,cMpc^{-1}}]$')
+            if normalize:
+                ax.set_ylabel(r'$\Delta^2 (k_{\bot})\,[\mathrm{%s}]$' % temp_unit)
+            else:
+                ax.set_ylabel(r'$P(k_{\bot})\,[\mathrm{%s\,h^{-3}\,cMpc^3}]$' % temp_unit)
+
+        if l_lambda:
+            x = self.el / (2 * np.pi)
+            ax.set_xlabel(r'$|\mathbf{u}|\,[\lambda]$')
+
+        if show68:
+            ax.fill_between(x, q16, q84, alpha=kargs.get('alpha', 0.5), color=kargs.get('c', None))
+        if show95:
+            ax.fill_between(x, q2_5, q97_5, alpha=0.5 * kargs.get('alpha', 0.25), color=kargs.get('c', None))
+
+        ax.plot(x, y, **kargs)
+
+        ax.set_yscale('log', **nonpos_arg)
+
+
+    def save(self, filename):
+        data = [p.data for p in self.all_ps]
+        err = [p.err for p in self.all_ps]
+        a = [self.freqs, data, err, self.el, self.k_per, self.w]
+        with h5py.File(filename, 'w') as hf:
+            for d, l in zip(a, ['freqs', 'data', 'err', 'el', 'k_per', 'ps_w']):
+                hf.create_dataset(l, data=d)
+
+    @staticmethod
+    def load(filename):
+        with h5py.File(filename, 'r') as hf:
+            data = hf.get('data')[:]
+            err = hf.get('err')[:]
+            freqs = hf.get('freqs')[:]
+            el = hf.get('el')[:]
+            k_per = hf.get('k_per')[:]
+            ps_w = hf.get('ps_w')
+            if not ps_w.size == 1:
+                ps_w = ps_w[:]
+
+        all_ps = [SpatialPowerSpectra(d, e, freqs, el, k_per) for d, e in zip(data, err)]
+
+        return VarianceMC(all_ps)
